"""
Unit-tests of the import and reloading functionality.
"""

import sys
import os
import shutil
import unittest
import tempfile
import errno

from figura import read_config
from figura.errors import ConfigParsingError
from figura.settings import get_setting, set_setting

################################################################################

# Tests use ".fig" extension. Enable it:
from figura.settings import set_extension_fig as _setfig
_setfig()

TEMPDIR_NAME = 'figura_%s'% os.getpid()
TEMPDIR = os.path.join(tempfile.gettempdir(), TEMPDIR_NAME)
BASEDIR = os.path.dirname(__file__)
CONFIGDIR = os.path.join(BASEDIR, 'config')
BASE_IMPORT_PATH = 'figura.tests.config'

################################################################################

class BasicTest(unittest.TestCase):

    #===================================================================================================================
    # setup / teardown
    #===================================================================================================================
    
    def setup_temp_dir(self, tempdir):
        for dir in [ TEMPDIR, tempdir ]:
            # create temp dir
            mkdir_p(dir)
            # touch __init__.py
            with open(os.path.join(dir, '__init__.py'), 'w'):
                pass

    def tearDown(self):
        # delete temp dir
        if os.path.isdir(TEMPDIR):
            shutil.rmtree(TEMPDIR)
    
    #===================================================================================================================
    # utility functions
    #===================================================================================================================
    
    def tempify_config(self, config_name, desc):
        config_name += '.%s' % get_setting('CONFIG_FILE_EXT')
        tempdir = os.path.join(TEMPDIR, desc)
        self.setup_temp_dir(tempdir)
        src = os.path.join(CONFIGDIR, config_name)
        dst = os.path.join(tempdir, config_name)
        shutil.copyfile(src, dst)
        return dst
    
    #===================================================================================================================
    # reloading
    #===================================================================================================================
    
    def test_reload_import_path(self):
        desc = 'test_reload_import_path'
        config_name = 'importee'
        filepath = self.tempify_config(config_name, desc)
        importpath = '.'.join([desc, config_name])
        self._test_reload(importpath, filepath)

    def test_reload_indirect_import_path(self):
        desc = 'test_reload_indirect_import_path'
        # "importer" imports "importee" -- loading "importer", but modifying "importee"
        filepath1 = self.tempify_config('importee', desc)
        self.tempify_config('importer', desc)
        importpath2 = '.'.join((desc, 'importer'))
        self._test_reload(importpath2, filepath1)

    # NOTE: loading by file-path is currently not supported
    
    #def test_reload_file_path(self):
    #    config_name = 'importee'
    #    filepath = self.tempify_config(config_name, 'test_reload_file_path')
    #    self._test_reload(filepath, filepath)
    #
    #def test_reload_indirect_file_path(self):
    #    # "importer" imports "importee" -- loading "importer", but modifying "importee"
    #    filepath1 = self.tempify_config('importee', 'test_reload_indirect_file_path')
    #    filepath2 = self.tempify_config('importer', 'test_reload_indirect_file_path')
    #    self._test_reload(filepath2, filepath1)

    def _test_reload(self, path_to_reload, filepath_to_modify):
        if TEMPDIR not in sys.path:
            sys.path.append(TEMPDIR)
        config = read_config(path_to_reload).some_params
        self.assertEqual(config.a, 1)
        self.assertRaises(KeyError, lambda: config['z'])
        self.assertRaises(AttributeError, lambda: config.z)
        self.assertRaises(AttributeError, lambda: config.zz)
        # inject a new value, and reload (indented, to put it inside some_params)
        append_line(filepath_to_modify, '    z = 999')
        config = read_config(path_to_reload).some_params
        self.assertEqual(config.a, 1)
        self.assertEqual(config.z, 999)
        self.assertRaises(AttributeError, lambda: config.zz)
        # inject yet another value, and reload (indented, to put it inside some_params)
        append_line(filepath_to_modify, '    zz = 555')
        config = read_config(path_to_reload).some_params
        self.assertEqual(config.a, 1)
        self.assertEqual(config.z, 999)
        self.assertEqual(config.zz, 555)
    
    #===================================================================================================================
    # extension
    #===================================================================================================================
    
    def test_same_name_import_vs_config1(self):
        path = '%s.samename' % BASE_IMPORT_PATH
        # read_config -- get the fig file
        self.assertEqual(read_config(path).filetype, 'fig')
        # import -- get the py file
        import tests.config.samename
        self.assertEqual(tests.config.samename.filetype, 'py')
    
    def test_same_name_import_vs_config2(self):
        # same as test_same_name_import_vs_config1, but order is reversed
        path = '%s.samename' % BASE_IMPORT_PATH
        # import -- get the py file
        import tests.config.samename
        self.assertEqual(tests.config.samename.filetype, 'py')
        # read_config -- get the fig file
        self.assertEqual(read_config(path).filetype, 'fig')
    
    def test_custom_extension(self):
        standard_path = '%s.basic1' % BASE_IMPORT_PATH
        custom_path = '%s.basic2' % BASE_IMPORT_PATH
        # standard: first works, second raises
        read_config(standard_path, enable_path_spliting = False)  # should not raise
        self.assertRaises(ConfigParsingError, read_config, custom_path, enable_path_spliting = False)
        # change to custom ext, then first raises, second works
        orig_ext = set_setting('CONFIG_FILE_EXT', 'customfig')
        try:
            self.assertRaises(ConfigParsingError, read_config, standard_path, enable_path_spliting = False)
            read_config(custom_path, enable_path_spliting = False)  # should not raise
        finally:
            set_setting('CONFIG_FILE_EXT', orig_ext)
        # change to a different ext, both should raise
        orig_ext = set_setting('CONFIG_FILE_EXT', 'nosuchfig')
        try:
            self.assertRaises(ConfigParsingError, read_config, standard_path, enable_path_spliting = False)
            self.assertRaises(ConfigParsingError, read_config, custom_path, enable_path_spliting = False)
        finally:
            set_setting('CONFIG_FILE_EXT', orig_ext)
        
def append_line(filename, line):
    with open(filename, 'a') as f:
        f.write('\n' + line + '\n')

def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise

################################################################################

if __name__ == '__main__':
    unittest.main()

################################################################################



# -*- coding: utf-8 -*-
import sys

__author__ = 'Kenneth M. "Elf" Sternberg'
__email__ = 'elf.sternberg@gmail.com'
__version__ = '0.1.0'

if sys.version_info[0] >= 3:
    from ._python3 import install, uninstall, is_installed, reset
elif sys.version_info[0:2] >= (2, 6):
    from ._python2 import install, uninstall, is_installed, reset

__all__ = ['install', 'uninstall', 'is_installed', 'reset']



"""
Running the self-contained unit-tests, i.e. config files containing
sections which are marked with is_unittest=True, which define what tests
to run on the configs in that file.
"""

import copy
import unittest

from figura import read_config, ConfigContainer

################################################################################

UNITTEST_FILE_PATH_PREFIX = 'figura.tests.config.'

# Tests use ".fig" extension. Enable it:
from figura.settings import set_extension_fig as _setfig
_setfig()

################################################################################

class BasicTest(unittest.TestCase):
    
    # make sure assertion messages include both the default message, and the
    # specialized message we pass along.
    longMessage = True
    
    #===================================================================================================================
    # utility functions
    #===================================================================================================================
    
    def assertConfigEqual(self, config1, config2, testname, key = 'ROOT'):
        #print(config1)
        if isinstance(config1, ConfigContainer) and isinstance(config2, ConfigContainer):
            self.assertSetEqual(
                set(config1.keys()), set(config2.keys()),
                msg = '[{%s} set of params at key=%s]' % (testname, key)
            )
            # recursively apply to sub containers
            for k in config1.keys():
                subkey = '%s.%s' % (key, k) if key else k
                self.assertConfigEqual(config1[k], config2[k], testname, key = subkey)
        else:
            self.assertEqual(config1, config2, '[{%s} value of param at key=%s]' % (testname, key))
    
    #===================================================================================================================
    # The main functions running the self-contained tests
    #===================================================================================================================
    
    def run_self_contained(self, path):
        # parse the figura file
        tests, data = self.parse(path)
        assert tests
        # run each test:
        for test in tests:
            #print('RUNNING %s / %s' % (path, test.name))
            self.run_single_test(test, data)

    def run_single_test(self, test, data):
        data = copy.deepcopy(data)
        
        # construct
        test_input = data
        entry_point = test.construct.get('entry_point')
        if entry_point:
            test_input = test_input.deep_getattr(entry_point)
        for base_path, ov_path in test.construct.get('apply_overrides', []):
            data.deep_getattr(base_path).apply_overrides(data.deep_getattr(ov_path))
        
        # verify
        test_output = data.deep_getattr(test.verify.entry_point)
        self.assertConfigEqual(test_input, test_output, test.name)
        
        # sanity: verify output is json-serializable
        test_output.to_json()

    def parse(self, path):
        # load the figura file
        data = read_config(UNITTEST_FILE_PATH_PREFIX + path)
        # extract test definitions
        tests = []
        for k, v in list(data.items()):
            if getattr(v, 'is_unittest', False):
                v.name = k
                tests.append(v)
                data.pop(k)
        return tests, data
        
    
    #===================================================================================================================
    # A test method per self-contained unittest
    #===================================================================================================================
    
    def test_basic1(self):
        return self.run_self_contained('basic1')
    def test_overlay(self):
        return self.run_self_contained('overlay')
    def test_override(self):
        return self.run_self_contained('override')
    

################################################################################

if __name__ == '__main__':
    unittest.main()

################################################################################



import os
import sys
import marshal
import pkgutil
import types
import _imp

if sys.version_info[0:2] in [(3, 3), (3, 4)]:
    from importlib._bootstrap import (cache_from_source, SourceFileLoader,
                                      FileFinder, _verbose_message, 
                                      _get_supported_file_loaders, _relax_case,
                                      _w_long, _code_type)


if sys.version_info[0:2] in [(3, 3)]:
    from importlib._bootstrap import _MAGIC_BYTES as MAGIC_NUMBER

if sys.version_info[0:2] == (3, 4):
    from importlib._bootstrap import _validate_bytecode_header, MAGIC_NUMBER

if sys.version_info[0:2] >= (3, 5):
    from importlib.machinery import SourceFileLoader, FileFinder
    from importlib._bootstrap import _verbose_message
    from importlib._bootstrap_external import (_w_long, _code_type, cache_from_source,
                                               _validate_bytecode_header,
                                               MAGIC_NUMBER, _relax_case,
                                               _get_supported_file_loaders)
    
SEP = os.sep
EXS = os.extsep
FLS = [('%s' + SEP + '__init__' + EXS + '%s', True), 
       ('%s' + EXS + '%s', False)]


def _suffixer(loaders):
    return [(suffix, loader) 
            for (loader, suffixes) in loaders
            for suffix in suffixes]


class _PolySourceFileLoader(SourceFileLoader):
    _compiler = None

    def _poly_bytes_from_bytecode(self, fullname, data, path, st):
        if hasattr(self, '_bytes_from_bytecode'):
            return self._bytes_from_bytecode(fullname, data,
                                             path, st)
        self_module = sys.modules[__name__]
        if hasattr(self_module, '_validate_bytecode_header'):
            return _validate_bytecode_header(data, source_stats = st,
                                             name = fullname, path = path)
        raise ImportError("No bytecode handler found loading.")

    # All this just to change one line.
    def get_code(self, fullname):
        source_path = self.get_filename(fullname)
        source_mtime = None
        try:
            bytecode_path = cache_from_source(source_path)
        except NotImplementedError:
            bytecode_path = None
        else:
            try:
                st = self.path_stats(source_path)
            except NotImplementedError:
                pass
            else:
                source_mtime = int(st['mtime'])
                try:
                    data = self.get_data(bytecode_path)
                except IOError:
                    pass
                else:
                    try:
                        bytes_data = self._poly_bytes_from_bytecode(fullname, data,
                                                                    bytecode_path,
                                                                    st)
                    except (ImportError, EOFError):
                        pass
                    else:
                        _verbose_message('{} matches {}', bytecode_path,
                                        source_path)
                        found = marshal.loads(bytes_data)
                        if isinstance(found, _code_type):
                            _imp._fix_co_filename(found, source_path)
                            _verbose_message('code object from {}',
                                            bytecode_path)
                            return found
                        else:
                            msg = "Non-code object in {}"
                            raise ImportError(msg.format(bytecode_path),
                                              name=fullname, path=bytecode_path)
        source_bytes = self.get_data(source_path)
        code_object = self._compiler(source_bytes, source_path, fullname)
        _verbose_message('code object from {}', source_path)
        if (not sys.dont_write_bytecode and bytecode_path is not None and
            source_mtime is not None):
            data = bytearray(MAGIC_NUMBER)
            data.extend(_w_long(source_mtime))
            data.extend(_w_long(len(source_bytes)))
            data.extend(marshal.dumps(code_object))
            try:
                self._cache_bytecode(source_path, bytecode_path, data)
                _verbose_message('wrote {!r}', bytecode_path)
            except NotImplementedError:
                pass
        return code_object
        

class PolyFileFinder(FileFinder):
    '''The poly version of FileFinder supports the addition of loaders
       after initialization.  That's pretty much the whole point of the 
       PolyLoader mechanism.'''

    _native_loaders = []
    _custom_loaders = []
    
    def __init__(self, path):
        # Base (directory) path
        self.path = path or '.'
        self._path_mtime = -1
        self._path_cache = set()
        self._relaxed_path_cache = set()

    @property
    def _loaders(self):
        return self._custom_loaders + list(self._native_loaders)
        
    @classmethod
    def _install(cls, compiler, suffixes):
        if not suffixes:
            return
        if isinstance(suffixes, str):
            suffixes = [suffixes]
        suffixset = set(suffixes)
        overlap = suffixset.intersection(set([suf[0] for suf in cls._native_loaders]))
        if overlap:
            raise RuntimeError("Override of native Python extensions is not permitted.")
        overlap = suffixset.intersection(
            set([loader[0] for loader in cls._custom_loaders]))
        if overlap:
            # Fail silently
            return

        newloaderclassname = (suffixes[0].lower().capitalize() + 
                              str(_PolySourceFileLoader).rpartition('.')[2][1:])
        if isinstance(compiler, types.FunctionType):
            newloader = type(newloaderclassname, (_PolySourceFileLoader,), 
                             dict(_compiler = staticmethod(compiler)))
        else:
            newloader = type(newloaderclassname, (_PolySourceFileLoader,), 
                             dict(_compiler = compiler))
        cls._custom_loaders += [(EXS + suffix, newloader) for suffix in suffixset]

    @classmethod
    def _uninstall(cls, suffixes):
        if not suffixes:
            return
        if isinstance(suffixes, str):
            suffixes = [suffixes]
        suffixset = set(suffixes)
        overlap = suffixset.intersection(set([suf[0] for suf in cls._native_loaders]))
        if overlap:
            raise RuntimeError("Removing a native Python extensions is not permitted.")

        exs_suffixset = set( EXS + suffix for suffix in suffixset )
        cls._custom_loaders = [ (suffix, loader) for suffix, loader in cls._custom_loaders if suffix not in exs_suffixset ]

    @classmethod
    def _is_installed(cls, suffix):
        suffix = EXS + suffix
        return any( suf == suffix for suf, loader in cls._custom_loaders )

    @classmethod
    def getmodulename(cls, path):
        filename = os.path.basename(path)
        suffixes = ([(-len(suf[0]), suf[0]) for suf in cls._native_loaders] +
                    [(-len(suf[0]), suf[0]) for suf in cls._custom_loaders])
        suffixes.sort()
        for neglen, suffix in suffixes:
            if filename[neglen:] == suffix:
                return filename[:neglen]
        return None

    def find_loader(self, fullname):
        """Try to find a loader for the specified module, or the namespace
        package portions. Returns (loader, list-of-portions)."""
        is_namespace = False
        tail_module = fullname.rpartition('.')[2]
        try:
            mtime = os.stat(self.path).st_mtime
        except OSError:
            mtime = -1
        if mtime != self._path_mtime:
            self._fill_cache()
            self._path_mtime = mtime
        # tail_module keeps the original casing, for __file__ and friends
        if _relax_case():
            cache = self._relaxed_path_cache
            cache_module = tail_module.lower()
        else:
            cache = self._path_cache
            cache_module = tail_module
        # Check if the module is the name of a directory (and thus a package).
        if cache_module in cache:
            base_path = os.path.join(self.path, tail_module)
            if os.path.isdir(base_path):
                for suffix, loader in self._loaders:
                    init_filename = '__init__' + suffix
                    full_path = os.path.join(base_path, init_filename)
                    if os.path.isfile(full_path):
                        return (loader(fullname, full_path), [base_path])
                else:
                    # A namespace package, return the path if we don't also
                    #  find a module in the next section.
                    is_namespace = True
        # Check for a file w/ a proper suffix exists.
        for suffix, loader in self._loaders:
            full_path = os.path.join(self.path, tail_module + suffix)
            _verbose_message('trying {}'.format(full_path), verbosity=2)
            if cache_module + suffix in cache:
                if os.path.isfile(full_path):
                    return (loader(fullname, full_path), [])
        if is_namespace:
            _verbose_message('possible namespace for {}'.format(base_path))
            return (None, [base_path])
        return (None, [])

    @classmethod
    def path_hook(cls, *loader_details):
        cls._native_loaders = loader_details
        def path_hook_for_PolyFileFinder(path):
            if not os.path.isdir(path):
                raise ImportError("only directories are supported", path=path)
            return PolyFileFinder(path)
        return path_hook_for_PolyFileFinder


def _poly_file_finder_modules(importer, prefix=''):
    if importer.path is None or not os.path.isdir(importer.path):
        return

    yielded = {}

    try:
        filenames = os.listdir(importer.path)
    except OSError:
        # ignore unreadable directories like import does
        filenames = []
    filenames.sort()
    for fn in filenames:
        modname = importer.getmodulename(fn)
        if modname == '__init__' or modname in yielded:
            continue

        path = os.path.join(importer.path, fn)
        ispkg = False

        if not modname and os.path.isdir(path) and '.' not in fn:
            modname = fn
            try:
                dircontents = os.listdir(path)
            except OSError:
                # ignore unreadable directories like import does
                dircontents = []
            for fn in dircontents:
                subname = importer.getmodulename(fn)
                if subname == '__init__':
                    ispkg = True
                    break
            else:
                continue    # not a package

        if modname and '.' not in modname:
            yielded[modname] = 1
            yield prefix + modname, ispkg
 

def install(compiler, suffixes):
    filefinder = [(f, i) for i, f in enumerate(sys.path_hooks)
                  if repr(f).find('.path_hook_for_FileFinder') != -1]
    if filefinder:
        filefinder, fpos = filefinder[0]
        sys.path_hooks[fpos] = PolyFileFinder.path_hook(*(_suffixer(_get_supported_file_loaders())))
        sys.path_importer_cache = {}
        pkgutil.iter_importer_modules.register(PolyFileFinder, _poly_file_finder_modules)

    PolyFileFinder._install(compiler, suffixes)

def uninstall(suffixes):
    PolyFileFinder._uninstall(suffixes)

def is_installed(suffix):
    return PolyFileFinder._is_installed(suffix)

def reset():
    PolyFileFinder._custom_loaders = []



"""
Tools for supporting config- and config-overrides-related CLI options
for python scripts.

This module provides integration with the ``argparse`` python CLI parser.
``optparse`` is currently not supported.

Example::

    import argparse
    from figura import read_config
    from figura.cli import add_override_argument
    # create the parser object:
    parser = argparse.ArgumentParser()
    # define cli arguments for specifying config file and overrides to apply to it:
    parser.add_argument('config', nargs = 1)
    add_override_argument(parser)
    # parse some arguments:
    args = parser.parse_args(['--override', 'x=somevalue', 'figura.tests.config.basic1'])
    # read the config and apply the overrides to it:
    config = read_config(args.config[0])
    config.apply_overrides(args.override)
    print(config)
    
Another example, leveraging `build_config <#figura.cli.build_config_from_cli>`__::

    import argparse
    from figura import read_config
    from figura.cli import add_override_argument
    # create the parser object:
    parser = argparse.ArgumentParser()
    # define cli arguments for specifying config file and overrides to apply to it:
    default_config = 'figura.tests.config.basic1'
    parser.add_argument('config', nargs = '*')  # 0 or more configs
    add_override_argument(parser)
    # parse some arguments:
    args = parser.parse_args(['--override', 'x=somevalue'])
    # read the config and apply the overrides to it:
    config = build_config_from_cli(args, default_config = default_config)
    print(config)

"""

import argparse
import six

from .container import ConfigContainer
from .utils import build_config

################################################################################
# argparse
################################################################################

DEFAULT_OVERRIDE_OPTION_NAMES = ( '-O', '--override' )


def add_override_argument(parser, *names, **kwargs):
    """
    Add to the argument-parser an argument for taking config-overrides via CLI.

    :note: The values of the overrides are read from CLI as strings. The
        code using them is responsible for converting them to other types
        as needed. Some convenience functions are provided for this purpose
        (e.g. boolify_).
    
    :param parser: an ArgumentParser object.
    :param names: The `names` args to be passed on to `parser.add_argument`.
        Defaults to `DEFAULT_OVERRIDE_OPTION_NAMES`.
    :param kwargs: Extra kwargs to pass on to `parser.add_argument`.
        Supported kw arguments: dest, required, help.

    .. _boolify:

    """
    if not names:
        names = DEFAULT_OVERRIDE_OPTION_NAMES
    dest = kwargs.pop('dest', None)
    required = kwargs.pop('required', False)
    help = kwargs.pop('help', 'extra overrides to apply to the config')
    if kwargs:
        raise TypeError('add_override_argument() got an invalid keyword argument: %s' % list(kwargs)[0])

    ov_container = ConfigContainer()
    ov_container.get_metadata().is_override_set = True
    parser.add_argument(
        *names,
        dest = dest,
        default = ov_container,
        required = required,
        action = _add_to_override_set,
        type = _dict_from_string,
        help = help
    )

def _dict_from_string(overrides_string):
    dct = {}
    for part in split_list_value(overrides_string):
        if not '=' in part:
            raise ValueError(part)
        k, _, v = part.partition('=')
        dct[k] = v
    return dct

class _add_to_override_set(argparse.Action):
    def __call__(self, parser, namespace, values, option_string):
        getattr(namespace, self.dest).update(values)

def build_config_from_cli(
        parsed_args,
        default_config = None,
        config_arg_name = 'config',
        override_arg_name = 'override',
        **kwargs):
    """
    A convenience function for combining configs and overrides from CLI.
    """
    configs = getattr(parsed_args, config_arg_name)
    overrides = getattr(parsed_args, override_arg_name)
    return build_config(
        *configs,
        extra_overrides = overrides,
        default_config = default_config,
        **kwargs
    )

################################################################################
# Escaping
################################################################################

OPTION_VALUE_ESCAPE_CHAR = '@'

def escape_list_value(value):
    """
    .. testsetup:: 
    
       from figura.cli import escape_list_value
       
    >>> escape_list_value('a')
    'a'
    >>> escape_list_value('a,b')
    'a@,b'
    >>> escape_list_value('@a@,b')
    '@@a@@@,b'
    """
    value = value.replace(OPTION_VALUE_ESCAPE_CHAR, OPTION_VALUE_ESCAPE_CHAR + OPTION_VALUE_ESCAPE_CHAR)
    value = value.replace(',', OPTION_VALUE_ESCAPE_CHAR + ',')
    return value

def split_list_value(value, delim = ','):
    """
    .. testsetup:: 
    
       from figura.cli import split_list_value
       
    >>> split_list_value('a,b,c')
    ['a', 'b', 'c']
    >>> split_list_value('a@,b@,c')
    ['a,b,c']
    >>> split_list_value('a@@,b@@,c')
    ['a@', 'b@', 'c']
    >>> split_list_value('@a,@b,@c@@')
    ['a', 'b', 'c@']
    >>> split_list_value('a,b,c@')
    ['a', 'b', 'c@']
    >>> split_list_value('')
    []
    """
    if not value:
        return []
    is_escaped = False
    tokens = ['']
    i = 0
    while i < len(value):
        c = value[i]
        if c == OPTION_VALUE_ESCAPE_CHAR and i < len(value) - 1:
            is_escaped = True
            i += 1
            c = value[i]
        else:
            is_escaped = False
        if c == delim and not is_escaped:
            tokens.append('')
        else:
            tokens[-1] += c
        i += 1
    return tokens

################################################################################
# Convenience functions for converting string values (from CLI) to other types
################################################################################

def _cast(v, t):
    if v is None:
        return None
    return t(v)

def intify(x):
    return _cast(x, int)

def floatify(x):
    return _cast(x, float)

_BOOLIFY_DICT = {
    0: False, 1: True,
    '0': False, '1': True,
    'no': False, 'yes': True,
    'n': False, 'y': True,
    'false': False, 'true': True,
    'disabled': False, 'enabled': True,
    'off': False, 'on': True,
}

def boolify(x):
    """
    Try to guess what boolean value ``x`` represents.
    
    :param x: could be a bool, string ("true", "yes", etc.), int (0 or 1)
    :raise ValueError: if can't determine bool-value from ``x``

    .. testsetup:: 
    
       from figura.cli import boolify
    
    >>> boolify(True)
    True
    >>> boolify(False)
    False
    >>> boolify('fAlse')
    False
    >>> boolify('yes')
    True
    >>> boolify(1)
    True
    >>> boolify(0)
    False
    
    """
    if isinstance(x, six.string_types):
        x = x.lower()
    try:
        return _BOOLIFY_DICT[x]
    except KeyError:
        raise ValueError('Can\'t boolify value: %r' % x)


################################################################################



"""
Useful tools when working with Figura configs.
"""

from .settings import get_setting
from .errors import ConfigError, ConfigParsingError
from .path import to_figura_path
from .container import ConfigContainer
from .parser import ConfigParser
from .importutils import walk_packages, is_importable_path, figura_implocked

################################################################################
# convenience functions

@figura_implocked
def read_config(path, enable_path_spliting = True):
    """
    Flexibly read/process a Figura config file.
    
    The path can point to:
    
    - a config file. E.g. ``figura.tests.config.basic1``
    - a config directory. E.g. ``figura.tests.config``
    - a value (or section) inside a config. E.g. ``figura.tests.config.basic1.some_params.a``
    
    :param path: a string or a `FiguraPath <#figura.path.FiguraPath>`_.
    :param enable_path_spliting: set to False if the path points to a file (as opposed to
        PATH.TO.FILE.PATH.TO.ATTR), if you want to suppress auto-splitting.
    :return: a `ConfigContainer <#figura.container.ConfigContainer>`_.
        In case of a deep path, the return value is the value from inside the
        conainer, which is not necessarilly a ConfigContainer.
        
    .. testsetup:: 

        from figura.utils import read_config
        
    >>> read_config('figura.tests.config.basic1').some_params.a  # read a config file
    1
    >>> read_config('figura.tests.config.basic1.some_params.a')  # read a value inside a config file
    1
    >>> read_config('figura.tests.config').basic1.some_params.a  # read a directory of config files
    1
    """
    
    # process the pass, split into file-path and attr-path
    file_path, attr_path = to_figura_path(path).split_parts()
    if not file_path:
        raise ConfigParsingError('No config file found for path: %r' % str(path))
    if not enable_path_spliting and attr_path:
        raise ConfigParsingError('No config file found for path: %r. In-file attributes detected: %s' % (
            str(path), attr_path))
    
    # parse the path:
    parser = ConfigParser()
    config = parser.parse(file_path)

    # support reading all modules under a package, and create a ConfigContainer
    # reflecting the structure:
    fig_ext = get_setting('CONFIG_FILE_EXT')
    for importer, rel_mod_path, ispkg in walk_packages(file_path):
        mod_path = '%s.%s' % ( file_path, rel_mod_path )
        if not ispkg and not is_importable_path(mod_path, with_ext = fig_ext):
            continue
        cur_config = parser.parse(mod_path)
        config.deep_setattr(rel_mod_path, cur_config)
    
    # apply the attr-path:
    if attr_path:
        config = config.deep_getattr(attr_path)

    return config


@figura_implocked
def build_config(*paths, **kwargs):
    """
    Build a configuration by reading Figura configs and optional
    override sets, and combining them into a final configuration.
    
    `read_config <#figura.utils.read_config>`_ is called for processing each path.

    :param paths: paths (strings or FiguraPaths) to config files. All but the first are
        treated as override-sets, and are applied to the base config.
        The first may also be an override-set, in which case,
        ``default_config`` must be specified, and is used as the base
        config.
    :param kwargs['default_config']: if the first path in ``paths`` is an overridet-set,
        default_config is used as the base config
    :param kwargs['extra_overrides']: a ConfigContainer of extra overrides to
        be applied to the config. ``extra_overrides`` are applied last.
    :param kwargs['enforce_override_set']:
        ensure that an override-sets is not used as base-config, and that a non-override-set
        is not used for overriding.
    :return: a `ConfigContainer <#figura.container.ConfigContainer>`_
    """

    default_config = kwargs.pop('default_config', None)
    extra_overrides = kwargs.pop('extra_overrides', None)
    enforce_override_set = kwargs.pop('enforce_override_set', True)
    if kwargs:
        raise TypeError('build_config() got an invalid keyword argument: %s' % list(kwargs)[0])
    
    configs = [ _to_config(conf) for conf in paths ]
    
    # using the default_config if the first config passed is an overrideset
    use_default = (len(configs) == 0) or \
        (isinstance(configs[0], ConfigContainer) and configs[0].get_metadata().is_override_set)
    if default_config is not None and use_default:
        configs = [ _to_config(default_config) ] + configs
    
    # read each config and combine them:
    is_first = True
    config = ConfigContainer()
    for cur_config in configs:
        if is_first:
            # This is the base config
            if enforce_override_set and cur_config.get_metadata().is_override_set:
                raise ConfigError('Attempting to use an override-set as a base config', cur_config)
            config = cur_config
            is_first = False
        else:
            # This is an override set to apply to the config
            config.apply_overrides(cur_config, enforce_override_set = enforce_override_set)
    if extra_overrides:
        config.apply_overrides(extra_overrides, enforce_override_set = enforce_override_set)
    return config

def _to_config(x):
    if isinstance(x, ConfigContainer):
        return x
    else:
        return read_config(x)

################################################################################



"""
A python package for parsing and working with `Figura <index.html>`_ config files.

This library also includes an executable ``figura_print`` command, for processing
your figura config files and generating your configuration, printing it as a JSON structure.
The generated JSON-structured configuration can then be read by programs written in
any other language (using its JSON parser).

In python code, simply use the `read_config <#figura.utils.read_config>`_ function for parsing
Figura config files into `ConfigContainer <#figura.container.ConfigContainer>`_ objects.
The `cli <#module-figura.cli>`_ module and the `build_config <#figura.utils.build_config>`_ function
are useful for integrating your python scripts with figura configurations and overrides, and combining
configurations and overrides into single ConfigContainer.

"""

from . import version

from .container import ConfigContainer
from .override import ConfigOverrideSet
from .errors import ConfigError, ConfigParsingError, ConfigValueError

from .utils import read_config, build_config



#! /usr/bin/env python
"""
Read a Figura config file, and print the config in JSON format.

Usage
------

By "python-import" path::
    
    % python figura_print.py figura.tests.config.basic1.some_params --override a=a_new_value

By file path: not support currently.

"""

import argparse

from figura import build_config, ConfigContainer
from figura.cli import add_override_argument

###############################################################################

FORMAT_MAP = dict(
    json = 'to_json',
    python = 'to_python_string',
)

###############################################################################
# MAIN

def main():
    args = getopt()
    paths = args.configfile + args.overridefile
    
    config = build_config(*paths, enforce_override_set = False)

    if args.override:
        config.apply_overrides(args.override)

    if isinstance(config, ConfigContainer):
        # format the container and print it:
        formatter_method = FORMAT_MAP[args.format]
        print(getattr(config, formatter_method)())
    else:
        # an atomic value. just print it:
        print(config)
            
###############################################################################

def getopt():

    parser = argparse.ArgumentParser(
        description='Read a figura file and print the configuration it defines')

    parser.add_argument('configfile', nargs = 1, help = \
                        '''a figura config file, or its python-import path''')

    parser.add_argument('overridefile', nargs = '*', help = \
                        '''figura files with override-sets, or their python-import path''')

    add_override_argument(parser)
    
    parser.add_argument('-f', '--format', default = 'json', choices = FORMAT_MAP.keys(),
                        help = '''The format of the output. defaults to 'json''')
    
    args = parser.parse_args()

    return args

###############################################################################

if __name__ == '__main__':
    main()



"""
Module-importing related tools, used for loading figura config files,
and inspecting packages containing them.
"""

import sys
import six
import functools
import imp
import importlib
import pkgutil
import figura.polyloader as polyloader  # TODO: figura.polyloader -> polyloader
try:
    # py3
    from importlib.util import find_spec as _find_module
except ImportError:
    # py2
    from pkgutil import find_loader as _find_module

from .settings import get_setting
from .errors import ConfigParsingError

################################################################################

class _NoImportSideEffectsContext(object):
    
    def __init__(self, cleanup_import_caches = True):
        self.cleanup_import_caches = cleanup_import_caches
        self._backup = {}
    
    def __enter__(self):
        # suppress writing of .pyc files:
        self.prev_dont_write_bytecode = sys.dont_write_bytecode
        sys.dont_write_bytecode = True
        
        if self.cleanup_import_caches:
            # remember which modules were already loaded before we run the import.
            self._backup_dict(sys.modules, 'modules', run_with_empty = False)
            self._backup_dict(sys.path_importer_cache, 'path_importer_cache', run_with_empty = True)
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.cleanup_import_caches:
            # remove all modules which got added to sys.modules in this import, to
            # ensure the next time we are here, they get reloaded.
            self._restore_dict(sys.modules, 'modules')
            self._restore_dict(sys.path_importer_cache, 'path_importer_cache')

        sys.dont_write_bytecode = self.prev_dont_write_bytecode
    
    def _backup_dict(self, dct, name, run_with_empty = False):
        if run_with_empty:
            self._backup[name] = ( True, dict(dct) )
            dct.clear()
        else:
            self._backup[name] = ( False, set(dct.keys()) )
    
    def _restore_dict(self, dct, name):
        run_with_empty, dct_backup = self._backup.pop(name)
        if run_with_empty:
            # restore dict values
            dct.clear()
            dct.update(dct_backup)
        else:
            # remove newly added keys
            for newkey in set(dct.keys()) - dct_backup:
                dct.pop(newkey, None)

class _FiguraImportContext(_NoImportSideEffectsContext):
    
    def __enter__(self):
        super(_FiguraImportContext, self).__enter__()
        self.should_uninstall = False
        fig_ext = get_setting('CONFIG_FILE_EXT')
        if fig_ext != '.py' and not polyloader.is_installed(fig_ext):
            polyloader.install(_figura_compile, fig_ext)
            self.should_uninstall = True
            self.fig_ext = fig_ext

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.should_uninstall:
            polyloader.uninstall(self.fig_ext)
        super(_FiguraImportContext, self).__exit__(exc_type, exc_val, exc_tb)


class _ImpLock(object):
    def acquire(self):
        imp.acquire_lock()
    def release(self):
        imp.release_lock()
    def __enter__(self):
        self.acquire()
    def __exit__(self, *a, **kw):
        self.release()

_implock = _ImpLock()

def figura_implocked(func):
    @functools.wraps(func)
    def f(*a, **kw):
        with _implock:
            with _FiguraImportContext():
                return func(*a, **kw)
    return f

def _figura_compile(source_bytes, source_path, fullname):
    """
    Just the standard python-file compiler (non-standard stuff might be added in the future).
    """
    return compile(source_bytes, source_path, 'exec', dont_inherit = True)

################################################################################

def _import_module_no_side_effects(path):
    """
    Similar to ``importlib.import_module(path)``, but with a few differences:
    
    - ``*.pyc`` files are not created as part of importing
    - the module imported isn't added to ``sys.modules``
    - calling this function multiple times with the same value will actually
      import the module multiple times (whereas standard importing caches the
      module in ``sys.modules``). this also applies for modules being imported
      indirectly during the importing of ``path``.
    """
    return _raw_import(path)

def _raw_import(path):
    # First, invalidate caches!
    # From the docs:
    # "If you are dynamically importing a module that was created since the interpreter began execution
    # (e.g., created a Python source file), you may need to call invalidate_caches() in order for the
    # new module to be noticed by the import system."
    # Since we're dealing with config files, we need to support re-reading config files
    # which are modified after the interpreter started.
    # See:
    # https://docs.python.org/3/library/importlib.html#importlib.import_module
    # https://docs.python.org/3/library/importlib.html#importlib.invalidate_caches
    try:
        invalidate_caches = importlib.invalidate_caches
    except AttributeError:
        # python2
        pass
    else:
        # python3
        invalidate_caches()
    # Now can safely import the module
    return importlib.import_module(path)

@figura_implocked
def import_figura_file(path):
    """
    Import a figura config file (with no side affects).
    
    :param path: a python import path
    :return: a python module object
    :raise ConfigParsingError: if importing fails
    """
    try:
        return _import_module_no_side_effects(path)
    except Exception as e:
        if six.PY2:
            # no exception chaining in python2
            #raise ConfigParsingError('Failed parsing "%s": %r' % (path, e)), None, sys.exc_info()[2]  # not a valid py3 syntax
            raise ConfigParsingError('Failed parsing config "%s": %r' % (path, e))
        else:
            #raise ConfigParsingError('Failed parsing config "%s"' % path) from e  # not a valid py2 syntax
            six.raise_from(ConfigParsingError('Failed parsing config "%s"' % path), e)

@figura_implocked
def is_importable_path(path, with_ext = None):
    """
    Does ``path`` point to a module which can be imported?

    :param path: a python import path
    """
    try:
        module_spec = _find_module(path)
    except (ImportError, AttributeError):
        module_spec = None
    if module_spec is None:
        return False
    if with_ext is not None:
        try:
            # py3
            filepath = module_spec.origin
        except AttributeError:
            try:
                # py2
                filepath = module_spec.get_filename()
            except AttributeError:
                # py2, but with polyloader
                filepath = module_spec.path
        if not filepath.endswith('.' + with_ext):
            return False
    return True

@figura_implocked
def walk_packages(file_path, prefix = '', skip_private = True):
    """
    Same as ``pkgutil.walk_packages``, except that it really does work recursively.
    """
    mod = import_figura_file(file_path)
    if hasattr(mod, '__path__'):
        
        # *PREFIX HACK*: for some reason, if we pass an empty prefix, walk_packages() can
        # yield packages not under the path we provide (this is probably a bug in walk_packages()).
        # E.g. if there's a "test" package under the __path__ passed, it can yield python's own
        # "test" package (e.g. (FileFinder('/usr/lib/python3.4/test'), 'test.pystone', False))
        # To bypass this bug, we make sure to always pass a non-empty prefix (and strip it back later).
        DUMMY_PREFIX = 'FIGURA___DUMMY___PREFIX.'
        tmp_prefix = DUMMY_PREFIX + prefix
        
        for importer, modname, ispkg in pkgutil.walk_packages(mod.__path__, prefix = tmp_prefix):

            # *PREFIX HACK* (continued)
            assert modname.startswith(DUMMY_PREFIX), modname
            modname = modname[len(DUMMY_PREFIX):]

            if skip_private and modname.startswith('_'):
                continue
            yield importer, modname, ispkg
            if ispkg:
                if prefix and modname.startswith(prefix):
                    modname = modname[len(prefix):]
                pref = '%s%s.' % (prefix, modname) if prefix else '%s.' % (modname,)
                for x in walk_packages('%s.%s' % (file_path, modname), prefix = pref):
                    yield x


################################################################################



"""
Global settings of figura
"""

from os import environ as _ENV
from .misc import Struct

################################################################################
# Defaults

DEFAULT_CONFIG_FILE_EXT = 'py'  # This is currently the default, in V1
#DEFAULT_CONFIG_FILE_EXT = 'fig'  # This will be enabled in V2.0 -- see set_extension_fig()

################################################################################
# The SETTINGS

SETTINGS = Struct(
    CONFIG_FILE_EXT = _ENV.get('FIGURA_CONFIG_FILE_EXT', DEFAULT_CONFIG_FILE_EXT),
)

################################################################################

def set_setting(key, new_value):
    old_val = SETTINGS[key]
    SETTINGS[key] = new_value
    return old_val

def get_setting(key, *default):
    return getattr(SETTINGS, key, *default)

################################################################################

def set_extension_fig():
    """
    Use this for forward-compatibility.  This enables the ".fig" extension
    (which will be the default in V2) while using Figura V1.
    
    Use like::
    
        from figura.settings import set_extension_fig; set_extension_fig()
     
    """
    return set_setting('CONFIG_FILE_EXT', 'fig')

################################################################################



"""
Definitions and tools of override-set-related and config-overriding-related operations.
"""

from .container import ConfigContainer
from .errors import ConfigError
from .misc import deep_getattr, deep_setattr
   
    
################################################################################

class ConfigOverrideSet(ConfigContainer):
    """
    A ConfigContainer representing an override-set, which can be applied to
    other config containers.
    """
    
    def __init__(self, *args, **kwargs):
        super(ConfigOverrideSet, self).__init__(*args, **kwargs)
        self.get_metadata().is_override_set = True

################################################################################

def apply_overrides_to_config(container, overrides, callback_key_prefix = '', enforce_override_set = True, **kwargs):
    """
    Apply overrides from an override-set to a container (in-place).
    
    :param container: a ``ConfigContainer`` to modify in-place
    :param overrides: a ``ConfigContainer`` containing the overrides to apply
    :param enforce_override_set: if ``enforce_override_set`` is set, raises an
        exception if ``overrides`` is not an override set (according to its
        ``is_override_set`` metadata attribute)
    :param kwargs: extra kwargs to pass on to `apply_override
        <#figura.override.apply_override>`_.
    :raise ConfigError: if ``enforce_override_set`` and ``overrides`` isn't an
        override set.
    """
    
    if enforce_override_set and not overrides.get_metadata().is_override_set:
        raise ConfigError('Attempting to apply overrides while is_override_set=False')
    
    for key, value in overrides.items():
        applied = False

        # check if overlaying -- i.e. if value is a nested container, which isn't
        # marked as opaque.
        old_value = container.get(key)
        if (isinstance(old_value, ConfigContainer) and
            isinstance(value, ConfigContainer) and
            not value.get_metadata().is_opaque_override):
            
            # old_value and value are both containers -- a nested override, apply recursively
            nested_container = container[key]
            cur_key_prefix = '%s.' % ( key, )
            if callback_key_prefix:
                cur_key_prefix = '%s.%s' % ( callback_key_prefix, cur_key_prefix )
            apply_overrides_to_config(
                nested_container, value,
                callback_key_prefix = cur_key_prefix,
                enforce_override_set = enforce_override_set,
                **kwargs
            )
            applied = True

        if not applied:
            # flat or plain override
            apply_override(
                container, key, value,
                callback_key_prefix = callback_key_prefix,
                **kwargs
            )
            applied = True

def apply_override(
        container, key, value,
        callback = None,
        callback_missing_value = '<<undef>>',
        callback_key_prefix = '',
        ):
    """
    Apply an override (key=value pair) to the given container in-place.
    The key can indicate override of a nested param, by including dots
    (or the alternative double-underscore form). E.g.: ``A.B.x``, or ``A__B__x``.
    
    :param container: a ConfigContainer to modify in place
    :param key: the attribute (or attr-path) to set
    :param value: the new attribute value
    :param callback: An optional callable to be called for "reporting" the
        override operation. The callable should take as args: key, value,
        old_value.
    :param callback_missing_value: the value to pass as old_value, in case the param
        was not previously set.
        
    """
    if callback is not None:
        old_value = deep_getattr(
            container, key, callback_missing_value,
            key_normalizer = normalize_override_key,
        )
        callback_key = callback_key_prefix + key
        callback(callback_key, value, old_value)
    deep_setattr(
        container, key, value,
        auto_constructor = type(container),
        key_normalizer = normalize_override_key,
    )

def normalize_override_key(key):
    """
    .. testsetup:: 

        from figura.override import normalize_override_key

    >>> normalize_override_key('a__b__c')
    'a.b.c'
    """
    if key.startswith('__'):
        return key
    return key.replace('__', '.')

################################################################################



"""
Definition of the config-container type.
"""

import json

from .misc import Struct, merge_dicts, deep_getattr, deep_setattr


################################################################################

class ConfigContainerMixin():
    """
    A mixin-baseclass of the `ConfigContainer <#figura.container.ConfigContainer>`_ class.
    """
    
    INDENT_LEN = 2
    
    DEFAULT_JSON_DUMP_KWARGS = {
        'indent': INDENT_LEN,
        'ensure_ascii': False,
    }

    #===================================================================================================================
    # attribute access
    #===================================================================================================================
    
    def deep_getattr(self, attr_path, *args):
        """
        See deep_getattr_.
        
        .. _deep_getattr:
        """
        return deep_getattr(self, attr_path, *args)

    def deep_setattr(self, attr_path, value):
        """
        See deep_setattr_.

        .. _deep_setattr:
        """
        return deep_setattr(self, attr_path, value)

    def apply_overrides(self, overrides, **kwargs):
        """
        A convenience method, simply calling apply_overrides_to_config_.
            
        :note: this method modifies the config container in-place.
        
        .. _apply_overrides_to_config: #figura.override.apply_overrides_to_config
        """
        from figura.override import apply_overrides_to_config  # avoid circular import
        apply_overrides_to_config(self, overrides, **kwargs)

    #===================================================================================================================
    # serialization
    #===================================================================================================================

    def to_string(self, **kwargs):
        """
        A JSON-string representation of the config container.
        """
        return self.to_json(**kwargs)
    
    def __str__(self):
        return self.to_string()
    
    def __repr__(self):
        return '%s(%s)' % (
            self.__class__.__name__,
            super(ConfigContainer, self).__repr__())
    
    def to_json(self, **kwargs):
        """
        :param args+kwargs: extra args to pass to ``json.dumps``.
        :return: the json string representation of self.
        """
        kw = dict(self.DEFAULT_JSON_DUMP_KWARGS)
        kw.update(kwargs)
        return json.dumps(self, **kw)
        
    @classmethod
    def from_json(cls, json_str):
        """
        Create a ConfigContainer from a json string.
        """
        return cls.from_dict(json.loads(json_str))
    
    def to_dict(self):
        """
        deep-conversion of self to a dict.
        """
        return json.loads(self.to_json())
    
    @classmethod
    def from_dict(cls, x):
        """
        deep-conversion of a dict to a ConfigContainer
        """
        return get_config_container_from_dict(x, cls = cls)

    def to_python_string(self):
        """
        A python-string representation of the config container.
        """
        lines = self._to_python_lines(self, 0)
        return '\n'.join(lines)
    
    def _to_python_lines(self, x, indent_level = 0):
        indent = ' ' * (self.INDENT_LEN * indent_level)
        lines = []
        for k, v in x.items():
            if isinstance(v, ConfigContainer):
                lines.append('%sclass %s:' % (indent, k))
                if len(v) > 0:
                    lines.extend(self._to_python_lines(v, indent_level + 1))
                else:
                    lines.append('%s%spass' % (indent, ' ' * self.INDENT_LEN))
            else:
                lines.append('%s%s = %r' % (indent, k, v))
        return lines

class ConfigContainer(Struct, ConfigContainerMixin):
    """
    A `Struct <#figura.misc.Struct>`_-like (recursive) container holding the configuration data.
    """
    
    DEFAULT_METADATA = Struct(
        is_override_set = False,
        is_opaque = False,
        is_opaque_override = False,
        doc = None,
    )
    
    def __init__(self, *args, **kwargs):
        metadata = kwargs.pop('metadata', self.DEFAULT_METADATA)
        super(ConfigContainer, self).__init__(*args, **kwargs)
        # not doing self._metadata=x because that results with self['_metadata']=x,
        # i.e. adding a new key to the container
        self.__dict__['_metadata'] = Struct(merge_dicts(self.DEFAULT_METADATA, metadata))
        
    def get_metadata(self):
        return self._metadata


################################################################################

def get_config_container_from_dict(x, cls = ConfigContainer):
    """
    Deep-conversion of a dict to a ConfigContainer.
    """
    if isinstance(x, dict):
        # convert recursively, and preserve dict type
        y = type(x)(
            ( k, get_config_container_from_dict(v, cls = cls) )
            for k, v in x.items()
        )
        return cls(y)
    else:
        # assume an atomic value
        return x

################################################################################



"""
Definition of the config-related exception types.
"""

################################################################################
# Exception types

class ConfigError(Exception):
    """
    Base class for config-related exceptions
    """
    pass

class ConfigParsingError(ConfigError):
    """
    Exception raised when configuration parsing fails
    """
    pass

class ConfigValueError(ConfigError):
    """
    Exception raised when configuration contains bad values or is missing values
    """
    pass

################################################################################



"""
Generally useful functions and classes used in this package.
"""

import copy
from collections import OrderedDict

################################################################################
# dict related

def merge_dicts(*args, **kwargs):
    """
    Merge dictionaries. Values in later args override former with the same key.
    kwargs always override.

    .. testsetup:: 

       from figura.misc import merge_dicts
    
    >>> sorted(merge_dicts({'a':1, 'b':2}, {'b':22, 'c':33, 'd':44}, d=444).items())
    [('a', 1), ('b', 22), ('c', 33), ('d', 444)]
     
    """
    if not args:
        args = ( {}, )
    d = copy.copy(args[0])
    for arg in args[1:]:
        
        # order-perserving: if type(arg)==OrderedDict, we assume the
        # caller wants to preserve the order of the items in the ordered dict
        if isinstance(arg, OrderedDict) and not isinstance(d, OrderedDict):
            d = OrderedDict(d)  # preserve order from now on
        
        d.update(arg)

    d.update(kwargs)
    return d

class Struct(dict):
    """
    A struct-like object. Can be thought of as a dict which also allows
    accessing its items using attribute-access::
    
        > x = Struct(a=3, b='bbb')
        > x
        {'a': 3, 'b': 'bbb'}
        > x.a
        3
        > x.a = 5
        > x.a
        5
        > x.c
        AttributeError: 'Struct' object has no attribute 'c'
        > x.c = 5
        > x.c
        5    
    """

    # a simpler implementation would use:
    #   self.__dict__ = self
    # but this can potentially leak memory under some python versions
    # (http://stackoverflow.com/questions/8687904/circular-referenced-objects-not-getting-garbage-collected)

    def __getattr__(self, k):
        try:
            return self[k]
        except KeyError:
            raise AttributeError(k)
    
    def __setattr__(self, k, v):
        self[k] = v        
    
    def __delattr__(self, k):
        del self[k]        
    
    def __dir__(self):
        return dir({}) + list(self.keys())

################################################################################
# attribute access tricks

def deep_getattr(x, attr_path, *args, **kwargs):
    """
    ``deep_getattr(x, 'a.b.c')`` -->
    ``getattr(getattr(getattr(x, 'a'), 'b'), 'c')``
    """
    try:
        return _deep_attr_operator(getattr, x, attr_path, *args, **kwargs)
    except AttributeError:
        # if a default value is provided, return it
        if args:
            default_value, = args
            return default_value
        raise

def deep_setattr(x, attr_path, value, **kwargs):
    """
    ``deep_setattr(x, 'a.b.c', y)`` -->
    ``setattr(getattr(getattr(x, 'a'), 'b'), 'c', y)``
    """
    return _deep_attr_operator(setattr, x, attr_path, value, **kwargs)

def _deep_attr_operator(func, x, attr_path, *args, **kwargs):
    auto_constructor = kwargs.pop('auto_constructor', None)
    key_normalizer = kwargs.pop('key_normalizer', lambda x: x)
    assert not kwargs, kwargs
    while True:
        attr, delim, rest = attr_path.partition('.')
        attr = key_normalizer(attr)
        if not delim:
            # deepest level -- apply func:
            return func(x, attr, *args)
        # go one level deeper:
        try:
            x = getattr(x, attr)
        except AttributeError:
            if auto_constructor is not None:
                setattr(x, attr, auto_constructor())
                x = getattr(x, attr)
            else:
                raise
        attr_path = rest

################################################################################



#!/usr/bin/env python

from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# Read version info from figura/version.py
version_vars = {}
with open("figura/version.py") as fp:
    exec(fp.read(), version_vars)
version_string = version_vars['__version_string__']

setup(
    name='figura',
    version=version_string,

    description='A python package for parsing and working with Figura config files',
    long_description=long_description,
    url='https://github.com/shx2/figura',
    author='shx2',
    author_email='shx222@gmail.com',
    license='MIT',

    packages=find_packages(exclude=['tests*', 'figura.tests*']),
    platforms = ["POSIX", "Windows"],
    install_requires=[
        'six',
    ],
    entry_points={
        'console_scripts': [
            'figura_print=figura.tools.figura_print:main',
        ],
    },

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Code Generators',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
    ],
    keywords='config, configuration, object oriented, code generation',

)



import io
import os
import os.path
import sys
import imp
import types
import pkgutil
from collections import namedtuple


SEP = os.sep
EXS = os.extsep
FLS = [('%s' + SEP + '__init__' + EXS + '%s', True), 
       ('%s' + EXS + '%s', False)]

Loader = namedtuple('Loader', 'suffix compiler')

class PolyLoader():
    def __init__(self, fullname, path, is_pkg):
        self.fullname = fullname
        self.path = path
        self.is_package = is_pkg

    def load_module(self, fullname):
        if fullname in sys.modules:
            return sys.modules[fullname]

        if fullname != self.fullname:
            raise ImportError("Load confusion: %s vs %s." % (fullname, self.fullname))

        matches = [loader for loader in PolyFinder._loader_handlers
                   if self.path.endswith(loader.suffix)]

        if len(matches) == 0:
            raise ImportError("%s is not a recognized module?" % fullname)

        if len(matches) > 1:
            raise ImportError("Multiple possible resolutions for %s: %s" % (
                fullname, ', '.join([loader.suffix for loader in matches])))

        compiler = matches[0].compiler
        with io.FileIO(self.path, 'r') as file:
            source_text = file.read()

        code = compiler(source_text, self.path, fullname)

        module = types.ModuleType(fullname)
        module.__file__ = self.path
        module.__name__ = fullname
        module.__package__ = '.'.join(fullname.split('.')[:-1])

        if self.is_package:
            module.__path__ = [os.path.dirname(module.__file__)]
            module.__package__ = fullname

        exec(code, module.__dict__)
        sys.modules[fullname] = module
        return module


# PolyFinder is an implementation of the Finder class from Python 2.7,
# with embellishments gleefully copied from Python 3.4.  It supports
# all the same functionality for non-.py sourcefiles with the added
# benefit of falling back to Python's default behavior.

# Polyfinder is instantiated by _polyloader_pathhook()

class PolyFinder(object):
    _loader_handlers = []
    _installed = False

    def __init__(self, path=None):
        self.path = path or '.'

    def _pl_find_on_path(self, fullname, path=None):
        subname = fullname.split(".")[-1]
        if self.path is None and subname != fullname:
            return None

        path = os.path.realpath(self.path)
        for (fp, ispkg) in FLS:
            for loader in self._loader_handlers:
                composed_path = fp % (('%s' + SEP + '%s') % (path, subname), loader.suffix)
                if os.path.isdir(composed_path):
                    raise IOError("Invalid: Directory name ends in recognized suffix")
                if os.path.isfile(composed_path):
                    return PolyLoader(fullname, composed_path, ispkg)

        # Fall back onto Python's own methods.
        try:
            file, filename, etc = imp.find_module(subname, [path])
        except ImportError as e:
            return None
        return pkgutil.ImpLoader(fullname, file, filename, etc)

    def find_module(self, fullname, path=None):
        return self._pl_find_on_path(fullname)

    @classmethod
    def _install(cls, compiler, suffixes):
        if isinstance(suffixes, basestring):
            suffixes = [suffixes]
        suffixes = set(suffixes)
        overlap = suffixes.intersection(set([suf[0] for suf in imp.get_suffixes()]))
        if overlap:
            raise RuntimeError("Override of native Python extensions is not permitted.")
        overlap = suffixes.intersection(
            set([loader.suffix for loader in cls._loader_handlers]))
        if overlap:
            # Fail silently
            return
        cls._loader_handlers += [Loader(suf, compiler) for suf in suffixes]

    @classmethod
    def _uninstall(cls, suffixes):
        if isinstance(suffixes, basestring):
            suffixes = [suffixes]
        suffixes = set(suffixes)
        overlap = suffixes.intersection(set([suf[0] for suf in imp.get_suffixes()]))
        if overlap:
            raise RuntimeError("Removing a native Python extensions is not permitted.")

        cls._loader_handlers = [ loader for loader in cls._loader_handlers if loader.suffix not in suffixes ]

    @classmethod
    def _is_installed(cls, suffix):
        return any( loader.suffix == suffix for loader in cls._loader_handlers )

    @classmethod
    def getmodulename(cls, path):
        filename = os.path.basename(path)
        suffixes = ([(-len(suf[0]), suf[0]) for suf in imp.get_suffixes()] +
                    [(-(len(suf[0]) + 1), EXS + suf[0]) for suf in cls._loader_handlers])
        suffixes.sort()
        for neglen, suffix in suffixes:
            if filename[neglen:] == suffix:
                return filename[:neglen]
        return None

    def iter_modules(self, prefix=''):
        if self.path is None or not os.path.isdir(self.path):
            return

        yielded = {}

        try:
            filenames = os.listdir(self.path)
        except OSError:
            # ignore unreadable directories like import does
            filenames = []
        filenames.sort()
        for fn in filenames:
            modname = self.getmodulename(fn)
            if modname == '__init__' or modname in yielded:
                continue

            path = os.path.join(self.path, fn)
            ispkg = False

            if not modname and os.path.isdir(path) and '.' not in fn:
                modname = fn
                try:
                    dircontents = os.listdir(path)
                except OSError:
                    # ignore unreadable directories like import does
                    dircontents = []
                for fn in dircontents:
                    subname = self.getmodulename(fn)
                    if subname == '__init__':
                        ispkg = True
                        break
                else:
                    continue    # not a package

            if modname and '.' not in modname:
                yielded[modname] = 1
                yield prefix + modname, ispkg


def _polyloader_pathhook(path):
    if not os.path.isdir(path):
        raise ImportError('Only directories are supported: %s' % path)
    return PolyFinder(path)


def install(compiler, suffixes):
    if not PolyFinder._installed:
        sys.path_hooks.append(_polyloader_pathhook)
        PolyFinder._installed = True
    PolyFinder._install(compiler, suffixes)

def uninstall(suffixes):
    if not PolyFinder._installed:
        return
    PolyFinder._uninstall(suffixes)

def is_installed(suffix):
    return PolyFinder._is_installed(suffix)

def reset():
    PolyFinder._loader_handlers = []
    PolyFinder._installed = False



"""
Parsing Figura config files.
"""

import inspect
import six

from .misc import merge_dicts
from .errors import ConfigParsingError
from .container import ConfigContainer
from .override import ConfigOverrideSet, normalize_override_key
from .importutils import import_figura_file


################################################################################
# Constants and definitions

VALID_ATOMIC_VALUE_TYPES = set([ type(None), type(''), type(u''), type(True), type(0), type(0.0) ])
VALID_SEQUENCE_VALUE_TYPES = set([ type(()), type([]), type({}) ])

def _add_type_of_exp(exp):
    try:
        VALID_ATOMIC_VALUE_TYPES.add(type(eval(exp)))
    except SyntaxError:
        pass

# in python2, also allow long:
_add_type_of_exp('0L')  # long

META_MAP = {
    '__override__': 'is_override_set',
    '__opaque__': 'is_opaque',
    '__opaque_override__': 'is_opaque_override',
    '__doc__': 'doc',
}

################################################################################
# The parser

class ConfigParser(object):
    """
    A python-config parser. Use like::
    
        from figura.parser import ConfigParser
        parser = ConfigParser()
        config = parser.parse('figura.tests.config.basic1')
        
    """
    
    VALID_ATOMIC_VALUE_TYPES = VALID_ATOMIC_VALUE_TYPES
    """ A list of atomic types allowed as config values """
    
    VALID_SEQUENCE_VALUE_TYPES = VALID_SEQUENCE_VALUE_TYPES
    """ A list of sequence types allowed as config values """
    
    META_MAP = META_MAP
    """
    A conversion map from config-metadata-directives (e.g. __opaque__),
    to their metadata attribute (e.g., is_opaque)
    """

    #===================================================================================================================
    # main parsing methods
    #===================================================================================================================

    def parse(self, path):
        """
        Parse a figura config file, identified by its import path, E.g.
        ``figura.tests.config.basic1``.
        
        :return: a `ConfigContainer <#figura.container.ConfigContainer>`_ object.
        :raise ConfigParsingError: if parsing fails.
        """
        
        # load it using python's import mechanism:
        x = self._import_python_module(path)
        # post-import processing:
        conf = self._python_to_conf(x)
        return conf

    #===================================================================================================================
    # python-import harnessing
    #===================================================================================================================
    
    def _import_python_module(self, path):
        """
        parse the python-config files into an intermediate representation. 
        """
        return import_figura_file(path)
    
    #===================================================================================================================
    # post-import processing of python constructs
    #===================================================================================================================

    def _python_to_conf(self, x, name = '', nesting_context = None, **metadata):
        """
        convert the intermediate config representation into a ConfigContainer. 
        """
        
        # If this is an atomic type, no parsing is required
        if type(x) in self.VALID_ATOMIC_VALUE_TYPES:
            return x
        elif type(x) in self.VALID_SEQUENCE_VALUE_TYPES:
            # a sequence. apply to elements recursively
            if hasattr(x, 'items'):
                return type(x)( (k, self._python_to_conf(v)) for k, v in x.items() )
            else:
                return type(x)( self._python_to_conf(elem) for elem in x )
        
        if type(x) == type(inspect):
            
            # the top-level module object --> a ConfigContainer
            raw_attrs = self._get_dunder_dict(x)
            
            # handle the __entry_point___ directive:
            new_x = self._find_entry_point(raw_attrs)
            if new_x is not None:
                # return the result of the recursive call on the new entry point
                return self._python_to_conf(new_x)

        elif _is_raw_container(x):
            # a class --> a ConfigContainer
            mro_dicts = [
                self._get_dunder_dict(t)
                for t in reversed(inspect.getmro(x))
            ]
            raw_attrs = merge_dicts(*mro_dicts)

            # handle auto-overlay magic
            if nesting_context is None:
                nesting_context = []
            nesting_context = nesting_context + [( name, x )]  # don't use append
            for overlayee in self._gen_overlayees(nesting_context):
                for k, v in self._get_dunder_dict(overlayee).items():
                    if k not in raw_attrs:
                        raw_attrs[k] = v
        else:
            # type misunderstood
            raise ConfigParsingError('Config construct of unsupported type %r: %s' % (type(x), x) )

        # Separate raw_attrs to their types. See _prepare_attrs for more details.
        real_attrs, meta_attrs = self._prepare_attrs(raw_attrs)
        
        # if this is an overridet-set, all nested containers are also override sets automatically.
        # we propagate this metadata attribute using the metadata dict:
        if meta_attrs.get('is_override_set', False):
            metadata['is_override_set'] = meta_attrs['is_override_set']

        # convert real attr recursively
        attrs = {
            k : self._python_to_conf(v, name = k, nesting_context = nesting_context, **metadata)
            for k, v in real_attrs.items()
        }
        
        # create the container
        metadata_to_apply = merge_dicts(metadata, meta_attrs)
        if metadata_to_apply.get('is_override_set', False):
            container_cls = ConfigOverrideSet
        else:
            container_cls = ConfigContainer
        container = container_cls(attrs)
        
        # set metadata attributes on the container
        container.get_metadata().update(metadata_to_apply)
        
        return container

    def _get_dunder_dict(self, x, deep = True):
        if x == object:
            # result of the mro call:
            return {}
        objs = [ x ]
        if deep:
            try:
                objs.extend(inspect.getmro(x)[1:])
            except AttributeError:
                pass
            dicts = [ self._get_dunder_dict(obj, deep = False) for obj in objs ]
        else:
            dicts = [ dict(x.__dict__) ]
        d = merge_dicts(*reversed(dicts))
        return d
    
    def _prepare_attrs(self, raw_attrs):
        """
        There are four types of attributes:
        
        1. real config attrs, e.g., a=1
        #. private config attrs, e.g., _a=1 -- we drop them
        #. metadata attrs, e.g., __opaque__=True
        #. python internal attrs, e.g., __module__=xxx -- we drop them
        
        :return: a 2-tuple: (real_attrs, meta_attrs)
        """
        real_attrs = {}
        meta_attrs = {}
        for k, v in raw_attrs.items():
            # try meta:
            try:
                meta_attrs[self.META_MAP[k]] = v
                continue
            except KeyError:
                pass  # not meta
            # filter out private attrs and python internals
            if k.startswith('_'):
                continue
            # the rest are real
            k = normalize_override_key(k)
            real_attrs[k] = v
        return real_attrs, meta_attrs
    
    #===================================================================================================================
    # overlay support
    #===================================================================================================================
    
    def _gen_overlayees(self, nesting_context):
        nesting_path = [ x[0] for x in nesting_context ]
        nesting_containers = [ x[1] for x in nesting_context ]
        rel_path = []
        while nesting_containers:
            nester = nesting_containers[-1]
            # discontinue overlay chain if __opaque__ is set in nester
            if _is_marked_opaque(nester):
                break
            if rel_path:
                for nester_base in inspect.getmro(nester)[1:]:
                    try:
                        attr_path = '.'.join(rel_path)
                        overlayee, opaque = \
                            self._deep_getattr_with_opaqueness_check(nester_base, attr_path)
                        # make sure the overlayee is a raw container, because it should
                        # be possible to override any (primitive) value with a container,
                        # in which case, no overlaying takes place.
                        if _is_raw_container(overlayee):
                            yield overlayee
                        # discontinue overlay chain if __opaque__ is set somehere
                        # in nester_base's path
                        if opaque:
                            break
                    except AttributeError:
                        pass
            rel_path = nesting_path[-1:] + rel_path
            nesting_path = nesting_path[:-1]
            nesting_containers = nesting_containers[:-1]
    
    def _deep_getattr_with_opaqueness_check(self, x, attr_path, *args):
        opaque = _is_marked_opaque(x)
        while True:
            attr, delim, rest = attr_path.partition('.')
            if delim:
                # go one level deeper:
                x = getattr(x, attr)
                attr_path = rest
                opaque = opaque or _is_marked_opaque(x)
            else:
                # deepest level
                x = getattr(x, attr_path, *args)
                opaque = opaque or _is_marked_opaque(x)
                return x, opaque

    #===================================================================================================================
    # __entry_point___ support
    #===================================================================================================================

    def _find_entry_point(self, attrs):
        entry_point = attrs.get('__entry_point__')
        if entry_point is not None:
            if isinstance(entry_point, six.string_types):
                # entry_point is given by name
                try:
                    entry_point = attrs[entry_point]
                except KeyError:
                    raise ConfigParsingError('entry-point not found: %s' % ( entry_point, ))
            else:
                # entry_point is given by pointing to to a container
                pass
        return entry_point

################################################################################

def _is_raw_container(x):
    # a raw container is the 'class x: ...' container as appears in the figura file, before
    # converting it to a ConfigContainer
    return inspect.isclass(x)

def _is_marked_opaque(x):
    return getattr(x, '__opaque__', False)

################################################################################



"""
Definition of `FiguraPath <#figura.path.FiguraPath>`_.
"""

from .settings import get_setting
from .importutils import is_importable_path

################################################################################

class FiguraPath(object):
    """
    A path to a Figura config file, or a value within one.
    The path contains two (optional) parts: the file-path, which is a python-import
    path (e.g. figura.tests), and a (deep) attribute-path to point to a specific
    value inside the file.
    
    The two parts are concatenated together with the same delimiter used by each
    of the parts, thus one cannot immediately tell
    
    E.g.: a path representing "the value of ``A.B.b`` inside the config file ``figura.tests.config.override``,
    is written like: ``figura.tests.config.override.A.B.b``.
    """
    
    DELIM = '.'
    """
    The delimiter used for all purposes: inside the file-path, between file-path
    and attr-path, and inside attr-path.
    """

    #===================================================================================================================
    
    def __init__(self, path):
        self._path = self._normalize_other(path)
    
    #===================================================================================================================
    # Path-parts handling
    #===================================================================================================================

    @classmethod
    def from_parts(cls, file_path, attr_path = None):
        """
        Construct a FiguraPath object from its two parts.
        
        :param file_path: a python import-path. E.g. ``figura.tests.config.override``
        :param attr_path: a (deep) attribute inside the file. E.g. ``A.B.b``
        """
        path_str = cls.DELIM.join([ part for part in (file_path, attr_path) if part ])
        return cls(path_str)

    def split_parts(self):
        """
        Split the path to its two parts: the file-path and the attr-path.
        Note this method looks for python modules in the filesystem to
        determine what prefix of the FiguraPath is a valid file-path. The
        rest is considered the attr-path.
        
        :return: a 2-tuple of (file_path, attr_path)
        
        .. testsetup:: 
    
            from figura.path import FiguraPath
        
        >>> FiguraPath('figura.hello_world.greeting.greetee').split_parts()
        ('figura.hello_world', 'greeting.greetee')
        """
        tokens = self._path.split(self.DELIM)
        n = len(tokens)
        for idx in reversed(range(n + 1)):
            file_path = self.DELIM.join(tokens[:idx])
            if self._is_figura_file_path(file_path):
                break
        return self.DELIM.join(tokens[:idx]), self.DELIM.join(tokens[idx:])

    def _is_figura_file_path(self, path):
        fig_ext = get_setting('CONFIG_FILE_EXT')
        return is_importable_path(path, fig_ext)
    
    #===================================================================================================================
    # operators for FiguraPath manipulation
    #===================================================================================================================

    def __add__(self, other):
        """
        Add a string to the end of the path.
        :note: this follows string semantics. No implicit delimiter is added.

        .. testsetup:: 
    
            from figura.path import FiguraPath
        
        >>> FiguraPath('abc') + 'd'
        FiguraPath('abcd')
        """
        other = self._normalize_other(other)
        return self._like_self(self._path + other)

    def __mod__(self, other):
        """
        The FiguraPath equivalent of string formatting.

        .. testsetup:: 
    
            from figura.path import FiguraPath
        
        >>> FiguraPath('foo.%s.bar') % 'baz'
        FiguraPath('foo.baz.bar')
        """
        other = self._normalize_other(other)
        return self._like_self(self._path % other)

    @property
    def parent(self):
        """
        The parent path, one level up.

        .. testsetup:: 
    
            from figura.path import FiguraPath
        
        >>> FiguraPath('a.bb.ccc').parent
        FiguraPath('a.bb')
        >>> FiguraPath('a').parent is None
        True
        """
        path = self._path.rpartition(self.DELIM)[0]
        if path:
            return FiguraPath(path)
        else:
            return None
    
    @property
    def basename(self):
        """
        The basename of the path.
        
        .. testsetup:: 
    
            from figura.path import FiguraPath

        >>> FiguraPath('a.b').basename
        'b'
        """
        x = self._path.rpartition(self.DELIM)[2]
        if x:
            return x
        else:
            return None
    
    def _normalize_other(self, other):
        try:
            return other._path
        except AttributeError:
            pass
        return other
    
    def _like_self(self, other):
        return type(self)(other)

    #===================================================================================================================
    # python stuff
    #===================================================================================================================

    def __repr__(self):
        return '%s(%r)' % ( type(self).__name__, self._path )
    
    def __str__(self):
        return self._path
    
    def __hash__(self):
        return hash(( type(self), str(self) ))
    
    def __eq__(self, other):
        try:
            return self._path == other._path
        except AttributeError:
            return NotImplemented
    
    def __ne__(self, other):
        return not (self == other)
    
    def __nonzero__(self):
        return bool(self._path)


def to_figura_path(path):
    """
    Convert values of different types to a FiguraPath.

    .. testsetup:: 

        from figura.path import FiguraPath, to_figura_path
    
    >>> to_figura_path('a.b.c')
    FiguraPath('a.b.c')
    >>> to_figura_path(FiguraPath('a.b.c'))
    FiguraPath('a.b.c')
    >>> to_figura_path(('a.b', 'C.d'))
    FiguraPath('a.b.C.d')
    """
    if isinstance(path, FiguraPath):
        return path
    elif isinstance(path, (list, tuple)) and len(path) == 2:
        return FiguraPath.from_parts(*path)
    else:
        return FiguraPath(path)

################################################################################
